/**
   This program takes an Abstract Syntax Tree (AST)
   that is defined by the following grammar, and
   converts it into a string that contains a
   C/C++/Java like infix notation.

       Prog ::= Exp
              | '(' 'prog' (Fun | Exp)+ Exp ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Exp ')'  // formal parameters followed by function body

        Exp ::= Array
              | IndexExp
              | Rand
              | For
              | IncDec
              | Apply
              | If
              | While
              | Begin
              | Set
              | Var
              | Print
              | AExp
              | BExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Array ::= '(' 'array' 'dim' Exp ')'       // instantiate an array value of size Exp
              | '(' 'array'  Exp+ ')'           // instantiate an array value literal
              | '(' 'array' 'dim' Exp Exp* ')'  // partially initialized array

   IndexExp ::= '(' 'index' Exp Exp ')'     // 1st operand is array, 2nd operand is integer

     SizeOf ::= '(' sizeOf Exp ')'          // operand must be an array

       Rand ::= '(' 'rand' Exp Exp ')'      // return a random integer between Exp and Exp

        For ::= '(' 'for' Exp Exp Exp Exp ')'

     IncDec ::= '(' '++'  VARIABLE ')'      // post-increment
              | '(' '--'  VARIABLE ')'      // post-decrement
              | '(' '+++' VARIABLE ')'      // pre-increment
              | '(' '---' VARIABLE ')'      // pre-decrement

      Apply ::= '(' 'apply' Exp Exp* ')'    // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'
              | '(' 'set' IndexExp Exp ')'  // set an array element

        Var ::= '(' 'var' VARIABLE Exp ')'

      Begin ::= '(' 'begin' Exp+ Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class AST2Infix_7a extends AST2Infix_7
{
   // Convert an expression
   protected String convertExp(Tree tree, String indent, boolean inline)
   {
      String result = "";

      String node = tree.getElement();

      if ( node.equals("array") )
         result += convertArray( tree, indent, inline );
      else if ( node.equals("index") )
         result += convertIndex( tree, indent, inline );
      else if ( node.equals("sizeOf") )
         result += convertSizeOf( tree, indent, inline );
      else if ( node.equals("rand") )
         result += convertRand( tree, indent, inline );
      else if ( node.equals("for") )
         result += convertFor( tree, indent, inline );
      else if ( node.equals("++")
             || node.equals("--")
             || node.equals("+++")
             || node.equals("---"))
         result += convertIncDec( tree, indent, inline );
      else
         result += super.convertExp( tree, indent, inline );

      return result;
   }//convertExp()


   // Convert an array expression
   protected String convertArray(Tree tree, String indent, boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += tree;  // just convert to an S-expression (CHANGE THIS)

      return result;
   }//convertArray()


   // Convert an index expression
   protected String convertIndex(Tree tree, String indent, boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += tree;  // just convert to an S-expression (CHANGE THIS)

      return result;
   }//convertIndex()


   // Convert a sizeOf expression
   protected String convertSizeOf(Tree tree, String indent, boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += tree;  // just convert to an S-expression (CHANGE THIS)

      return result;
   }//convertSizeOf()


   // Convert a rand expression
   protected String convertRand(Tree tree, String indent, boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += tree;  // just convert to an S-expression (CHANGE THIS)

      return result;
   }//convertRand()


   // Convert a for-loop expression
   protected String convertFor(Tree tree, String indent, boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += tree;  // just convert to an S-expression (CHANGE THIS)

      return result;
   }//convertFor()


   // Convert an increment/decrement expression
   protected String convertIncDec(Tree tree, String indent, boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += tree;  // just convert to an S-expression (CHANGE THIS)

      return result;
   }//convertIncDec()

}//AST2Infix_7a